import numpy as np
import matplotlib.pyplot as plt
import torch

class DNN(torch.torch.nn.Module):
  def __init__(self, xsize):
    super(DNN, self).__init__()
    self.xsize=xsize

    self._convsize=1024
    self.fc_front=torch.nn.Sequential(
        torch.nn.Linear(xsize, self._convsize),
        torch.nn.ReLU(),
        torch.nn.Dropout(0.5)
    )
    self.fc1=torch.nn.Sequential(
        torch.nn.Conv1d(1, 1, 512, 1, "same"),
        torch.nn.ReLU()
    )
    self.fc2=torch.nn.Sequential(
        torch.nn.Conv1d(1, 1, 128, 1, "same"),
        torch.nn.ReLU()
    )
    self.fc3=torch.nn.Sequential(
        torch.nn.Conv1d(1, 1, 32, 1, "same"),
        torch.nn.ReLU()
    )
    self.fc4=torch.nn.Sequential(
        torch.nn.Conv1d(1, 1, 8, 1, "same"),
        torch.nn.ReLU()
    )
    self.fc5=torch.nn.Sequential(
        torch.nn.Conv1d(1, 1, 2, 1, "same"),
        torch.nn.ReLU()
    )
    self.fc_rear=torch.nn.Sequential(
        torch.nn.Linear(5*self._convsize, 128),
        torch.nn.ReLU(),
        torch.nn.Dropout(0.5),
        torch.nn.Linear(128, 3)
    )
  
  def forward(self, x):
    y1=self.fc_front(x.view(-1, self.xsize))
    y2=torch.cat((self.fc1(y1), self.fc2(y1), self.fc3(y1), self.fc4(y1), self.fc5(y1)), dim=1)
    y3=self.fc_rear(y2)
    return y3

def data(I, k, beta):
  global x
  return I*(np.sin(k*x)/k/x)**2*(np.cos(beta*x))**2

if __name__=="__main__":
  x_size=1000; TrainN=100000; validN=50;
  x=np.linspace(-1e-2, 1e-2, x_size)
  model=DNN(x_size)
  optim=torch.optim.Adam(model.parameters(), lr=1e-3)
  lossF=torch.nn.MSELoss()
  model.train()
  losses=[]

  I0=np.random.uniform(0, 1e2, validN)
  k0=np.random.uniform(0, 1e3, validN)
  beta0=np.random.uniform(0, 1e4, validN)

  for i in range(TrainN):
    print("\r Trained: %d/%d" %(i+1, TrainN), end="", flush=True)
    I=np.random.uniform(0, 1e2)
    k=np.random.uniform(0, 1e3)
    beta=np.random.uniform(0, 1e4)
    label=torch.tensor([I, k, beta]).view(1, 3)
    attr=torch.tensor(data(I, k, beta), dtype=torch.float32)
    optim.zero_grad()
    pred=model(attr)
    loss=lossF(label, pred)
    loss.backward()
    optim.step()

    # Validation
    if(i%100==0):
      model.eval()
      loss_sum=0
      for j in range(validN):
        label_valid=torch.tensor([I0[j], k0[j], beta0[j]]).view(1, 3)
        attr_valid=torch.tensor(data(I0[j], k0[j], beta0[j]), dtype=torch.float32)
        pred=model(attr_valid)
        loss=lossF(label_valid, pred)
        loss_sum+=loss.detach()
      losses.append(loss_sum/validN)

  
  print("\n\n")
  plt.plot(losses, color="#ff00ff")
  plt.title("Valid Loss")
  plt.yscale("log"); plt.ylabel(f"$Validation Loss[Log]$")
  plt.show()
